/*^
 * Module:       grc_fix.c
 * ```````
 * Overview:     Bitmap image transpose utility.
 * `````````
 * Originated:   2015  M.J.Bauer  [www.mjbauer.biz/mjb_resources.htm]
 * ```````````
 *
 * This module implements a command-line utility named 'grcfix', the purpose of which is to
 * transpose bitmap image definitions from Microchip 'GRC' format into a raw bitmap format
 * designed for use with MJB's graphics LCD function library -- LCD_graphics_lib.
 *
 * Microchip Graphics Resources Converter (GRC) utility may be downloaded from Microchip's
 * developer website. It forms part of the (legacy) Microchip Library Applications (MLA) for
 * the PIC32MX MCU family. GRC takes as input any common image format, e.g. JPEG, BMP, TIFF,
 * and outputs image data in a C source file (array definitions, etc) suitable for compilation
 * into an embedded application. The bitmap image data format output from GRC is intended for
 * use with Microchip's LCD graphics library functions, e.g. PutImage(). In the case of 1-bit-
 * per-pixel images, pixels are ordered right-to-left in each byte, i.e. bit7 is the rightmost
 * pixel and bit0 is the leftmost.  MJB's LCD graphics function, LCD_PutImage(), expects the
 * order of pixels in a byte to be left-to-right, i.e. bit7 is the leftmost pixel, because
 * this format results in more efficient (faster) image rendering.
 *
 * This utility, grcfix, takes as its input bitmap image data definitions generated by GRC,
 * stored in MCU flash program memory, and outputs image data in a C source file (array
 * definitions, etc) suitable for compilation into an embedded application which will use
 * MJB's LCD graphics library functions. Images to be converted by GRC must be created using
 * 1-bit-per-pixel (black & white) data format. An image editor (Photoshop, MS Paint, etc)
 * can be used to convert the data format to 1BPP.  Note: GRC outputs 1's for white, 0's for
 * black in the colour palette, i.e. white is the foreground colour, black is background.
 *
 * To use the utility, the GRC source file containing the image definitions to be transposed
 * must be re-named "grc_resources.c". This file, plus the matching header file "grc_resources.h"
 * must be included in the "grcfix" application project. When compiled and loaded into the
 * target PIC32MX MCU, the program will have all of the image data from "grc_resources.c"
 * accessible in flash memory. The source code of the command function 'Cmnd_grcfix()' must
 * be edited and re-compiled to select the desired image objects to be transposed and output.
 * The transposed image definitions will be output to 'stdout' (CLI serial port) when the
 * command 'grcfix' is executed.
 *
 * Remove the following line from files "grc_resources.c" and "grc_resources.h":
 * #include <Graphics/Graphics.h>
 *
 * Add the following line to file "grc_resources.c":
 * #include "grc_defs.h"
 *
 * ================================================================================================
 */
#include "system_def.h"
#include "grc_defs.h"
#include "grc_resources.h"
#include "console.h"


/*
 * Function reverses the order of bits in the byte passed as argument.
 */
uint8  ReverseOrderBits(uint8 bDat)
{
    int    i;
    uint8  revByte = 0x00;

    for (i = 0;  i < 8;  i++)
    {
        revByte <<= 1;
        if (bDat & 1) revByte |= 0x01;
        bDat >>= 1;
    }

    return  revByte;
}


/*
 * CLI command function:  Cmnd_grcfix
 *
 * The command "grcfix" transforms bitmap image definitions generated by Microchip's "GRC"
 * application into bitmap image definitions for use with LCD_graphics_lib functions.
 *
 * Command syntax:  grcfix <opt>
 */
void  Cmnd_grcfix(int argCount, char * argValue[])
{
    IMAGE_FLASH *imageDesc;
    char   imageName[64];
    char   lineBuf[128];
    uint8  bDat;
    uint8 *pData;
    int    imageSize, imageWidth, imageHeight;
    int    i, column = 0;
    char   option = tolower(argValue[1][1]);

    // Set up a pointer to the GRC image descriptor to be processed.
    // The image must be defined in the source file, grc_resources.c.
    // Output imageName must not be the same as input from grc_resources.c.
    //
    imageDesc = (IMAGE_FLASH *) &Bauer_remi_logo;     // defined in grc_resources.c
    imageSize =  Bauer_remi_logo_SIZE;                // defined in grc_resources.h
    strcpy(imageName, "Bauer_remi_logo_85x45");       // image name (string) for output

    //======= The following code is the same for every image to be processed =======

    if (argCount == 1 || *argValue[1] == '?')   // help wanted
    {
        putstr( "Usage:  grcfix  {-i|-n} \n" );
        putstr( "GRC bitmap transpose utility.\n" );
        putstr( "If <opt> is '-i', invert GRC image bits; else don't.\n" );
        return;
    }

    imageSize -= GRC_IMAGE_HEADER_SIZE_1BPP;
    imageHeight = *((uint16 *)(imageDesc->address) + 1);
    imageWidth =  *((uint16 *)(imageDesc->address) + 2);
    pData = (uint8 *)(imageDesc->address) + GRC_IMAGE_HEADER_SIZE_1BPP;

    putstr("\n\n/* \n");
    putstr(" * Bitmap image definition \n");
    sprintf(lineBuf, " * Image name: %s, width: %d, height: %d pixels \n",
            imageName, imageWidth, imageHeight);
    putstr(lineBuf);
    putstr(" */ \n");

    sprintf(lineBuf, "bitmap_t  %s[] = \n{ ", imageName);
    putstr(lineBuf);

    for (i = 0;  i < imageSize;  i++, pData++)
    {
        if ((column % 16) == 0) putstr("\n    ");

        bDat = ReverseOrderBits(*pData);
        if (option == 'i') bDat ^= 0xFF;  // invert bits
        sprintf(lineBuf, "0x%02X", bDat);
        putstr(lineBuf);
        if (i < (imageSize - 1)) putstr(", ");

        if (++column >= 16)  column = 0;
    }

    putstr("\n}; \n\n");
}
