/***************************************************************************************************
 * File:      keypad_4wire.c
 *
 * Overview:  API and driver functions for 16-button (4 x 4) numeric keypad,
 *            using 4-wire analog interface to the MCU.
 *
 * Note:      If the ADC module is to be accessed by other tasks, the keypad service routine
 *            may be suspended temporarily by the other task using the function
 *            KeypadTaskEnable(DISABLE), then re-activated after the other task is done.
 *
 * Originated:   2015  M.J. Bauer   [www.mjbauer.biz]
 *
 */
#include "system_def.h"
#include "keypad_4wire.h"
#include "console.h"

#ifdef USE_4X4_KEYPAD

static  volatile BOOL   keypadTaskActive;       // flag
static  volatile short  keypadTaskState;        // current state of ISR task
static  volatile BOOL   keyHitDetected;         // flag: key hit detected (de-glitched)
static  volatile BOOL   keyHitRegistered;       // flag: key hit registered
static  volatile short  lastHitKeyCol;          // column of key last hit (0..3)
static  volatile short  lastHitKeyRow;          // row of key last hit (0..3)
static  volatile short  KeycodeLastHit;         // ASCII-encoded value of last key hit
static  volatile long   KeyPressTime_ms;        // Duration of key press (ms)

// Current raw ADC readings (unit = ADC count)
static  volatile short  ADCreadingKeyCol;       // Keypad col ADC reading (raw)
static  volatile short  ADCreadingKeyRow;       // Keypad row ADC reading (raw)
static  volatile short  ADCreadingTrash;        // Dummy reading


/***************************************************************************************************
 * Function:     void  KeypadTaskEnable()
 *
 * Overview:     Enables or disables the foreground task, Keypad_Service().
 *               The task must be disabled if the ADC module is to be used to read
 *               analog input signals other than those supported by the task.
 *
 * Entry arg's:  (BOOL) enab = TRUE to enable the ADC task, FALSE to disable it.
 */
void  KeypadTaskEnable(BOOL enab)
{
    keypadTaskActive = 0;

    if (enab)
    {
        keypadTaskState = KEY_SENSE_INIT;
        keypadTaskActive = 1;
    }
}


/***************************************************************************************************
 * Function:     uint8 KeyHit()
 *
 * Overview:     Tests for key hit, i.e. transition from not pressed to pressed.
 *               An internal flag, keyHitRegistered, is cleared on exit so that the function
 *               will return TRUE once only for each key press event.
 *
 * Entry arg's:  --
 *
 * Return val:   TRUE (1) if a key hit was detected since the last call, else FALSE (0).
 */
uint8  KeyHit(void)
{
    uint8 result = 0;

    if (keyHitRegistered)
    {
        result = 1;
        keyHitRegistered = 0;
    }

    return result;
}


/***************************************************************************************************
 * Function:     void  WaitForKeyHit()
 *
 * Overview:     Waits indefinitely for a key to be pressed, then returns.
 *               While waiting, any active background tasks are executed.
 *               On return, the keycode of the key hit can be obtained by a call
 *               to function GetKey().
 *
 * Caution:      WaitForKeyHit() must not be called from any background task
 *               dispatched by BackgroundTaskExec(), for obvious reasons.
 *
 */
void  WaitForKeyHit(void)
{
    while (!keyHitRegistered)
    {
        BackgroundTaskExec();   // Function defined in main or R/T kernel module

#if USE_CONSOLE_CLI
        ConsoleCLI_Service();   // Function defined in console CLI module
#endif
    }

    keyHitRegistered = 0;
}


/***************************************************************************************************
 * Function:     short GetKey()
 *
 * Overview:     Returns the ASCII keycode of the last key hit detected, i.e.
 *               following a call to function KeyHit() which returned TRUE.
 *
 * Entry arg's:  --
 *
 * Return val:   keycode (ASCII), range: '0'..'9', 'A', 'B', 'C', 'D', '*', '#'.
 */
uint8  GetKey(void)
{
    return  KeycodeLastHit;
}


/*********************************************************************
* Function:      short  GetKeyPressTime_ms()
*
* Overview:      Returns time (ms) for which a key is held down.
*                Returns zero if no key pressed.
*
********************************************************************/
short  GetKeyPressTime_ms(void)
{
    return  KeyPressTime_ms;
}


/*********************************************************************
* Function:      short  GetADCreadingKeyCol()
*
* Overview:      Returns raw ADC reading of keypad column input
*                while a key is being pressed.
*                Returns (-1) if no key pressed.
*
********************************************************************/
short  GetADCreadingKeyCol(void)
{
    return  ADCreadingKeyCol;
}


/*********************************************************************
* Function:      short  GetADCreadingKeyCol()
*
* Overview:      Returns raw ADC reading of keypad row input
*                while a key is being pressed.
*                Returns (-1) if no key pressed.
*
********************************************************************/
short  GetADCreadingKeyRow(void)
{
    return  ADCreadingKeyRow;
}


/*********************************************************************
* Function:      void  FindKeycode()
*
* Overview:      Calculates ASCII code of last key hit from
*                row and column ADC input readings.
*
********************************************************************/
static  void  FindKeycode(void)
{
    static  const  uint8  keycode_LUT[][4] =
            { '1', '2', '3', 'A',
              '4', '5', '6', 'B',
              '7', '8', '9', 'C',
              '*', '0', '#', 'D'
            };

    short  row = 3, col = 3;   // row and column index values (0..3)

    if (ADCreadingKeyCol < 170) col = 0;  else
    if (ADCreadingKeyCol < 500) col = 1;  else
    if (ADCreadingKeyCol < 850) col = 2;

    if (ADCreadingKeyRow < 170) row = 0;  else
    if (ADCreadingKeyRow < 500) row = 1;  else
    if (ADCreadingKeyRow < 850) row = 2;

    KeycodeLastHit = keycode_LUT[row][col];
}


/**************************************************************************************************
 * Function:  KeypadService(void)
 *
 * Overview:  Service Routine for 4-wire keypad (4 x 4 buttons).
 *            The keypad handler is very similar to that used for a 4-wire resistive touch-screen.
 *
 * Detail:    Called periodically (at 1ms intervals) from Timer 1 (RTI) ISR.
 *            When a keypad "hit" is detected, the function determines the row and
 *            column, hence the key-code, of the keypad button which is pressed.
 *            Keypad service total cycle time is 5ms (assuming 1ms call interval).
 *
 * Note:      If the ADC module is to be accessed by other tasks, the keypad service routine
 *            must be suspended temporarily by the other task using the function call
 *            KeypadTaskEnable(DISABLE), then re-enabled after the other task is done.
 *
 *************************************************************************************************/
void  KeypadService(void)
{
    static short  tempX, tempY;
    static short  last_state = 0;
    static short  watchdog_count = 0;  // number of calls in same state
    static short  key_down = 0;

    if (key_down) KeyPressTime_ms++;  // a key is being pressed

    if (!keypadTaskActive)  return;

    switch (keypadTaskState)
    {
        case KEY_SENSE_INIT:
        {
            // Initialize the ADC for manual sampling mode
            AD1CON1 = 0;            // reset
            AD1CON2 = 0;            // AVdd, AVss, MUXA only
            AD1CON3 = 0x0002;       // Tad = 4 * Tcy
            AD1CON1 = 0x8000;       // Turn on A/D module, use manual convert
            AD1CSSL = 0;            // No scanned inputs

            // Setup for "key down" sensing (on digital input)
            XPLUS_DRIVE_LOW();          // Output 0V on X+
            XPLUS_CONFIG_AS_OUTPUT();   //   ..   ..
            XMINUS_DRIVE_LOW();         // Output 0V on X-
            XMINUS_CONFIG_AS_OUTPUT();  //   ..   ..

            YPLUS_CONFIG_AS_INPUT();    // Y+ is digital input (100k pullup)
            YMINUS_CONFIG_AS_INPUT();   // float Y-

            ADPCFG_PIN_XPLUS = PIN_CONFIG_DIGITAL;
            ADPCFG_PIN_XMINUS = PIN_CONFIG_DIGITAL;
            ADPCFG_PIN_YPLUS = PIN_CONFIG_DIGITAL;
            ADPCFG_PIN_YMINUS = PIN_CONFIG_DIGITAL;

            keypadTaskState = KEY_DOWN_SENSE;
            break;
        }
        case KEY_DOWN_SENSE:
        {
            TESTPOINT_KEYISR_HIGH();    // Scope trigger

            if (YPLUS_INPUT_STATE() == 0)  // LOW => key is pressed
            {
                key_down = 1;
            }
            else  // HIGH => no key pressed
            {
                key_down = 0;
                keyHitDetected = 0;
                KeyPressTime_ms = 0;
            }

            // Setup to sample X voltage (key column)
            ADC_INPUT_SEL = ADC_CHAN_YMINUS;   // X voltage is read on Y- input

            YMINUS_CONFIG_AS_INPUT();    // ADC input Y- (X voltage)
            YPLUS_CONFIG_AS_INPUT();     // float Y+

            XMINUS_DRIVE_LOW();         // Output 0V on X-
            XMINUS_CONFIG_AS_OUTPUT();

            XPLUS_DRIVE_HIGH();         // Output +3.3V on X+
            XPLUS_CONFIG_AS_OUTPUT();   //

            ADPCFG_PIN_XMINUS = PIN_CONFIG_DIGITAL;  // X- set to digital pin
            ADPCFG_PIN_XPLUS = PIN_CONFIG_DIGITAL;   // X+ set to digital pin
            ADPCFG_PIN_YMINUS = PIN_CONFIG_ANALOG;   // Y- set to analog pin

            ADC_SAMPLING = 1;    // start sampling
            keypadTaskState = KEY_X_HOLD;
            break;
        }
        case KEY_X_HOLD:
        {
            TESTPOINT_KEYISR_LOW();
            
            ADC_SAMPLING = 0;    // stop sampling (hold & start conversion)
            keypadTaskState = KEY_X_READ;
            break;
        }
        case KEY_X_READ:
        {
            if (!ADC_CONV_DONE) break;  // wait for conversion

            tempX = ADC_RESULT_REG;     // Read X voltage

            // Setup to sample Y voltage (key column)
            ADC_INPUT_SEL = ADC_CHAN_XMINUS;   // Y voltage is read on X- input

            XMINUS_CONFIG_AS_INPUT();    // ADC input X- (Y voltage)
            XPLUS_CONFIG_AS_INPUT();     // float X+

            YMINUS_DRIVE_LOW();         // Output 0V on Y-
            YMINUS_CONFIG_AS_OUTPUT();

            YPLUS_DRIVE_HIGH();         // Output +3.3V on Y+
            YPLUS_CONFIG_AS_OUTPUT();

            ADPCFG_PIN_YMINUS = PIN_CONFIG_DIGITAL;   // Y- set to digital pin
            ADPCFG_PIN_YPLUS = PIN_CONFIG_DIGITAL;    // Y+ set to digital pin
            ADPCFG_PIN_XMINUS = PIN_CONFIG_ANALOG;    // X- set to analog pin

            ADC_SAMPLING = 1;    // start sampling
            keypadTaskState = KEY_Y_HOLD;
            break;
        }
        case KEY_Y_HOLD:
        {
            ADC_SAMPLING = 0;    // stop sampling (hold & start conversion)
            keypadTaskState = KEY_Y_READ;
            break;
        }
        case KEY_Y_READ:
        {
            if (!ADC_CONV_DONE) break;  // wait for conversion

            tempY = ADC_RESULT_REG;     // Read Y voltage

            if (!keyHitDetected && (KeyPressTime_ms > KEY_HIT_DETECT_TIME))
            {
                keyHitDetected = 1;   // prevent unwanted repeat hits
                keyHitRegistered = 1;
                ADCreadingKeyCol = tempX;   // capture ADC readings at time of hit
                ADCreadingKeyRow = tempY;
                FindKeycode();              // convert ADC readings to key code
            }

            // Setup for "key down" sensing (on digital input)
            XPLUS_DRIVE_LOW();          // Output 0V on X+
            XPLUS_CONFIG_AS_OUTPUT();   //   ..   ..
            XMINUS_DRIVE_LOW();         // Output 0V on X-
            XMINUS_CONFIG_AS_OUTPUT();  //   ..   ..

            YPLUS_CONFIG_AS_INPUT();    // Y+ is digital input (100k pullup)
            YMINUS_CONFIG_AS_INPUT();   // float Y-

            ADPCFG_PIN_XPLUS = PIN_CONFIG_DIGITAL;
            ADPCFG_PIN_XMINUS = PIN_CONFIG_DIGITAL;
            ADPCFG_PIN_YPLUS = PIN_CONFIG_DIGITAL;
            ADPCFG_PIN_YMINUS = PIN_CONFIG_DIGITAL;

            keypadTaskState = KEY_DOWN_SENSE;
            break;
        }
        default:
            keypadTaskState = KEY_SENSE_INIT;
            break;
    }

    // Trap failure and recover in case ADC routine gets stuck in same state
    if (keypadTaskState == last_state)
    {
        if (++watchdog_count > 10)  // allow up to 10 iterations in same state
        {
            watchdog_count = 0;
            keypadTaskState = KEY_SENSE_INIT;
        }
    }

    last_state = keypadTaskState;
}


/************************************************************************************
 * Function:     short ADC_Service_GetState()
 *
 * Overview:     Returns the ADC service routine state, e.g. for diagnostics.
 *
 */
short  Keypad_GetState(void)
{
    return keypadTaskState;
}

#endif  // USE_4X4_KEYPAD
