/**
 * Module:       pcm_audio.c
 *
 * Overview:     PCM audio player functions using PIC32MX OCx for PWM output.
 *
 * Originated:   M.J.Bauer, 2014   [www.mjbauer.biz]
 *
 * This module provides functions to play sound effects from PCM audio data files
 * stored in MCU internal flash PM or in external SPI flash memory.
 *
 * Reference:    "Synthesis of Digital (PCM) Sound Effects", M.J.Bauer (2014)
 *
 * Note:         This module is not intended for inclusion in the "REMI" controller project;
 *               it is incompatible with the "REMI" controller application.
 *
 * ================================================================================================
 */
#include <sys/attribs.h>    // For interrupt handlers

#include "pcm_audio.h"
#include "data_flash_fs.h"
#include "console.h"

void  SystemHeartbeatLEDtask(void);

float  g_AudioVolume;      // audio volume control, user-settable param (log, 0..100)

static volatile int    v_MasterLevel;   // master audio level (full scale = 16384 = 2^14)
static volatile uint32 v_PlayTimer;
static volatile uint32 v_PlayDuration;  // clip length, samples
static volatile uint8  v_PlayTrigger;
static volatile uint8  v_PlayActive;
static volatile uint8  v_PlayMode;

static int16  *pSoundSample;    // Pointer to internal flash sound data


/**
 * Function:     Initialize Timer_2 and OC#/RD# for PWM output generation.
 *
 * Timer_2 is set up with Prescaler = 1:1;  Fclk = FCY = 80MHz;  Tclk = 12.5ns
 * Timer_2 period := 50.00us (4000 x 12.5ns);  PR2 = 4001;  PWM freq := 20MHz.
 * Maximum duty register value is 3999.
 *
 * Configure OCx channel for PWM, fault detect disabled. (See pcm_audio.h)
 */
void  PCM_Audio_Init(void)
{
    TRISD = TRISD & ~(1 << PORTD_BIT);  // RDx/OCx is an output pin
    OC_CON_REG = 0x0000;                // Disable OCx while timer is being set up

    OpenTimer2(T2_ON | T2_PS_1_1, 4001);
    ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_4);

    OC_RS_REG = OC_R_REG = 2000;        // Set initial duty (50%)
    OC_CON_REG = 0x8006;                // Enable OCx channel

    g_AudioVolume = 70;
}


/**
 * Function:     Start playback of sound clip^ from PCM data array stored in internal
 *               MCU flash program memory.  (^Typically very short clips only.)
 *
 * Entry arg:    soundDefn = address of sound definition (WORD array) in program flash.
 *               clipLength = number of samples to be played
 */
void  PlaySound(const int16 *soundDefn, uint32 clipLength)
{
    v_PlayActive = 0;   // Stop any sound currently playing

    pSoundSample = (int16 *) soundDefn;
    v_MasterLevel = g_AudioVolume * g_AudioVolume * 1.6;  // Maximum 16000
    v_PlayDuration = clipLength;
    v_PlayMode = PLAY_SOUND_INTERNAL_FLASH;
    v_PlayTrigger = 1;    // Let 'er rip, Boris.
}


/**
 * Function:     Start playback of PCM sound file (or part thereof) stored in external
 *               flash memory (SDF file).
 *
 * Entry arg:    file_ID = PCM sound file ID number (SDF directory index)
 *               clipLength = number of samples to be played
 *
 * Note:         Background tasks are suspended while the audio player is active, because
 *               the player ISR locks up the SPI channel used for SDF memory access.
 *               Otherwise, tasks using the SPI channel could be disrupted by the player.
 */
void  PlaySoundFile(short file_ID, uint32 clipLength)
{
#ifdef USE_SDF_FILE_SYSTEM

    if (v_PlayActive)
    {
        v_PlayActive = 0;
        if (v_PlayMode == PLAY_SOUND_EXTERNAL_FLASH)  SDF_fclose();
    }

    if (SDF_fopen_RD(file_ID))  // file opened OK...
    {
        v_MasterLevel = g_AudioVolume * g_AudioVolume * 1.6;  // Maximum 16000
        v_PlayDuration = clipLength;
        v_PlayMode = PLAY_SOUND_EXTERNAL_FLASH;
        v_PlayTrigger = 1;    // Let 'er rip, Boris.
    }

    do   // Wait for player to stop, while keeping the front-panel LED alive
    {
        Delay_ms(50);
        SystemHeartbeatLEDtask();
    }
    while (v_PlayActive);

    SDF_fclose();
#endif
}


/**
 * Function:     Find PCM sound file by name in SDF memory and start playback.
 *               The playing time (clipLength) is inferred from the file size.
 *
 * Entry arg:    sdf_filename = SDF sound clip filename (string, up to 19 chars)
 */
void  PlaySoundFileNamed(char *sdf_filename)
{
#ifdef USE_SDF_FILE_SYSTEM

    DataFlash_FileIndexRecord_t *pFileInfo;
    int  file_ID;
    uint32  clipLength;

    if ((file_ID = SDF_find(sdf_filename)) != ERROR)
    {
        pFileInfo = SDF_finfo(file_ID);
        clipLength = pFileInfo->fileSize / 2;  // # of samples
        PlaySoundFile(file_ID, clipLength);
    }
#endif
}


/**
 * Function:     Return TRUE if the PCM audio player is active.
 */
BOOL  AudioPlayerActive(void)
{
    return  (v_PlayActive != 0);
}


/**
 * Function:     Timer_2 OVF interrupt service routine -- PCM audio driver
 */
void  __ISR(_TIMER_2_VECTOR, IPL5AUTO)  Timer_2_IRQService(void)
{
    static int16  sample;

    TestpointOutputRA7(1);   // 'Scope signal -- to measure ISR execution time

    if (v_PlayTrigger)  // (re-)start audio clip
    {
        v_PlayTimer = 0;
        v_PlayActive = 1;
        v_PlayTrigger = 0;
    }

    if (v_PlayActive)
    {
#ifdef USE_SDF_FILE_SYSTEM

        if (v_PlayMode == PLAY_SOUND_EXTERNAL_FLASH)
        {
            sample  = (uint16) SDF_fread() & 0xFF;    // LSB
            sample += (uint16) SDF_fread() << 8;      // MSB
        }
        else
#endif
        if (v_PlayMode == PLAY_SOUND_INTERNAL_FLASH)
        {
            sample = *pSoundSample++;
        }
        else  v_PlayActive = 0;  // Play mode undefined

        if (++v_PlayTimer >= v_PlayDuration) v_PlayActive = 0;  // reached end-of-clip
    }
    else sample = 0;

    sample = ((int32) sample * v_MasterLevel) >> 18;   // Scale and apply master volume control
    OC_RS_REG = sample + 2000;                         // Add bias (50%) before writing

    TestpointOutputRA7(0);
	IFS0bits.T2IF = 0;         // Clear the IRQ
}


// PCM sound clip data, length: 1600 samples, duration: 80 ms
// OSC1 : Sine (1), freq= 2500 Hz, level= 1.000, offset= 0.000
// ENV  : holdoff= 0 ms, attack= 2 ms, sustain= 60 ms, release= 10 ms
//
const int16  short_beep_2500Hz[] =
    {
         0,      0,    750,   1060,      0,  -2121,  -3750,  -3181,      0,   4242,
      6750,   5303,      0,  -6363,  -9750,  -7424,      0,   8485,  12750,   9545,
         0, -10606, -15750, -11667,      0,  12727,  18749,  13788,      0, -14849,
    -21749, -15909,      0,  16970,  24749,  18031,      0, -19091, -27749, -20152,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29999,  21213,      0, -21213, -29999, -21213,      0,  21213,
     29999,  21213,      0, -21213, -29999, -21213,      0,  21213,  29999,  21213,
         0, -21213, -29999, -21213,      0,  21213,  29999,  21213,      0, -21213,
    -29999, -21213,      0,  21213,  29999,  21213,      0, -21213, -29999, -21213,
         0,  21213,  29849,  21001,      0, -20792, -29257, -20584,      0,  20379,
     28676,  20176,      0, -19974, -28107, -19775,      0,  19578,  27549,  19383,
         0, -19189, -27002, -18998,      0,  18808,  26466,  18621,      0, -18435,
    -25941, -18251,      0,  18069,  25426,  17889,      0, -17710, -24921, -17534,
         0,  17359,  24426,  17186,      0, -17014, -23941, -16844,      0,  16676,
     23466,  16510,      0, -16345, -23000, -16182,      0,  16021,  22544,  15861,
         0, -15703, -22096, -15546,      0,  15391,  21658,  15238,      0, -15086,
    -21228, -14935,      0,  14786,  20806,  14639,      0, -14493, -20393, -14348,
         0,  14205,  19988,  14063,      0, -13923, -19592, -13784,      0,  13646,
     19203,  13510,      0, -13376, -18822, -13242,      0,  13110,  18448,  12979,
         0, -12850, -18082, -12722,      0,  12595,  17723,  12469,      0, -12345,
    -17371, -12222,      0,  12100,  17026,  11979,      0, -11859, -16688, -11741,
         0,  11624,  16357,  11508,      0, -11393, -16032, -11280,      0,  11167,
     15714,  11056,      0, -10945, -15402, -10836,      0,  10728,  15096,  10621,
         0, -10515, -14797, -10410,      0,  10306,  14503,  10204,      0, -10102,
    -14215, -10001,      0,   9901,  13933,   9803,      0,  -9705, -13656,  -9608,
         0,   9512,  13385,   9417,      0,  -9323, -13119,  -9230,      0,   9138,
     12859,   9047,      0,  -8957, -12604,  -8867,      0,   8779,  12353,   8691,
         0,  -8605, -12108,  -8519,      0,   8434,  11868,   8350,      0,  -8266,
    -11632,  -8184,      0,   8102,  11401,   8022,      0,  -7941, -11175,  -7862,
         0,   7784,  10953,   7706,      0,  -7629, -10736,  -7553,      0,   7478,
     10523,   7403,      0,  -7329, -10314,  -7256,      0,   7184,  10109,   7112,
         0,  -7041,  -9908,  -6971,      0,   6902,   9712,   6833,      0,  -6765,
     -9519,  -6697,      0,   6630,   9330,   6564,      0,  -6499,  -9145,  -6434,
         0,   6370,   8963,   6306,      0,  -6243,  -8785,  -6181,      0,   6119,
      8611,   6058,      0,  -5998,  -8440,  -5938,      0,   5879,   8272,   5820,
         0,  -5762,  -8108,  -5704,      0,   5648,   7947,   5591,      0,  -5535,
     -7789,  -5480,      0,   5426,   7635,   5371,      0,  -5318,  -7483,  -5265,
         0,   5212,   7335,   5160,      0,  -5109,  -7189,  -5058,      0,   5007,
      7046,   4957,      0,  -4908,  -6906,  -4859,      0,   4811,   6769,   4763,
         0,  -4715,  -6635,  -4668,      0,   4621,   6503,   4575,      0,  -4530,
     -6374,  -4484,      0,   4440,   6248,   4395,      0,  -4352,  -6124,  -4308,
         0,   4265,   6002,   4223,      0,  -4181,  -5883,  -4139,      0,   4098,
      5766,   4057,      0,  -4016,  -5652,  -3976,      0,   3936,   5539,   3897,
         0,  -3858,  -5429,  -3820,      0,   3782,   5322,   3744,      0,  -3707,
     -5216,  -3670,      0,   3633,   5112,   3597,      0,  -3561,  -5011,  -3525,
    };

